// setdns.cpp : definisce il punto di ingresso dell'applicazione console.
//

#include "stdafx.h"

#ifdef _UNICODE
std::wostream &tcout = std::wcout;
std::wostream &tcerr = std::wcerr;
#define tstring wstring
#else
std::ostream &tcout = std::cout;
std::ostream &tcerr = std::cerr;
#define tstring string
#endif

#define LENGTHOF(x) (sizeof(x) / sizeof((x)[0]))

typedef int (CALLBACK* DNSFLUSHPROC)();
typedef int (CALLBACK* DHCPNOTIFYPROC)(LPWSTR, LPWSTR, BOOL, DWORD, DWORD, DWORD, int);

void usage()
{
	tcout <<
		_T("usage: setdsn [options] <interface>\n")
		_T("options:\n")
		_T("  -d, --dhcp              : set DNS to DHCP\n")
		_T("  -s, --static <address>  : set DNS to the specified address\n")
		_T("      --help              : show this help page\n")
	<< std::endl;
}

bool IsValidIp(LPTSTR szAddress)
{
#ifdef _UNICODE
	char szAddressA[32];
	if (!WideCharToMultiByte(CP_ACP, 0, szAddress, -1, szAddressA, sizeof(szAddressA), NULL, NULL))
		return false;
#else
	char *szAddressA = szAddress;
#endif
	pcre *re;
	const char *error;
	int erroffset;
	int ovector[30];

	re = pcre_compile(
		"^(?:(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])\\.){3}(?:25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9]?[0-9])$",
		0,
		&error,
		&erroffset,
		NULL
	);

	bool res = pcre_exec(re, NULL, szAddressA, strlen(szAddressA), 0, 0, ovector, LENGTHOF(ovector)) >= 0;
	pcre_free(re);

	return res;
}

bool IsDHCPEnabled(LPCTSTR szAdapter)
{
	HKEY hKey;
	std::tstring strKeyName = _T("SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\\");
	strKeyName += szAdapter;

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, strKeyName.c_str(), 0, KEY_READ, &hKey) != ERROR_SUCCESS)
		return false;
	
	DWORD dwEnabled = 0;
	DWORD cbBuffer = sizeof(dwEnabled);
	bool res = RegQueryValueEx(hKey, _T("EnableDHCP"), NULL, NULL, (LPBYTE)&dwEnabled, &cbBuffer) == ERROR_SUCCESS &&
		dwEnabled == 1;

	RegCloseKey(hKey);

	return res;
}

bool RegSetDNS(LPCTSTR szAdapter, LPCTSTR szDNS)
{
	HKEY hKey;
	std::tstring strKeyName = _T("SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\\");
	strKeyName += szAdapter;

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, strKeyName.c_str(), 0, KEY_WRITE, &hKey) != ERROR_SUCCESS)
		return false;
	
	size_t len = szDNS
		? _tcslen(szDNS) + 1
		: 1;

	LPTSTR pDNS = szDNS
		? szDNS
		: _T("\0");

	bool res = RegSetValueEx(hKey, _T("NameServer"), 0, REG_SZ, (LPCBYTE)pDNS, len * sizeof(TCHAR)) == ERROR_SUCCESS;

	RegCloseKey(hKey);

	return res;
}

bool GetAdapterName(LPCTSTR szAdapter, LPTSTR szBuffer, DWORD cbBuffer)
{
	HKEY hKey;
	std::tstring strKeyName = _T("SYSTEM\\CurrentControlSet\\Control\\Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}\\");
	strKeyName += szAdapter;
	strKeyName += _T("\\Connection");

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, strKeyName.c_str(), 0, KEY_READ, &hKey) != ERROR_SUCCESS)
		return false;

	bool res = RegQueryValueEx(hKey, _T("Name"), NULL, NULL, (LPBYTE)szBuffer, &cbBuffer) == ERROR_SUCCESS;

	RegCloseKey(hKey);

	return res;
}

bool NotifyIPChange(LPWSTR szAdapter)
{
	bool res = false;
	HINSTANCE hDhcpDll;
	DHCPNOTIFYPROC pDhcpNotifyProc;
	
	if ((hDhcpDll = LoadLibrary(_T("dhcpcsvc"))) == NULL)
		return false;

	if ((pDhcpNotifyProc = (DHCPNOTIFYPROC)GetProcAddress(hDhcpDll, "DhcpNotifyConfigChange")) != NULL)
		if ((pDhcpNotifyProc)(NULL, szAdapter, FALSE, 0, NULL,NULL, 0) == ERROR_SUCCESS)
			res = true;

	FreeLibrary(hDhcpDll);

	return res;
}

bool FlushDNS()
{
	bool res = false;
	HINSTANCE hDnsDll;
	DNSFLUSHPROC pDnsFlushProc;

	if((hDnsDll = LoadLibrary(_T("dnsapi"))) == NULL)
		return false;

	if ((pDnsFlushProc = (DNSFLUSHPROC)GetProcAddress(hDnsDll, "DnsFlushResolverCache")) != NULL)
		if ((pDnsFlushProc)())
			res = true;
	
	FreeLibrary(hDnsDll);

	return res;
}

bool SetDns(LPCTSTR szAdapterName, LPCTSTR szDNS)
{
	ULONG ulAdapterInfoSize = 0;
	IP_ADAPTER_INFO *pInfoHead = NULL;

	if (GetAdaptersInfo(NULL, &ulAdapterInfoSize) != ERROR_BUFFER_OVERFLOW)
		return false;

	IP_ADAPTER_INFO *pAdapterInfo = (IP_ADAPTER_INFO *)new char[ulAdapterInfoSize];
	pInfoHead = pAdapterInfo;

	if (GetAdaptersInfo(pAdapterInfo, &ulAdapterInfoSize) != ERROR_SUCCESS)
	{
		delete[] pInfoHead;
		return false;
	}

	do
	{
#ifdef _UNICODE
		wchar_t szTAdapter[MAX_ADAPTER_NAME_LENGTH + 4];
		if (MultiByteToWideChar(CP_ACP, 0, pAdapterInfo->AdapterName, sizeof(pAdapterInfo->AdapterName), szTAdapter, LENGTHOF(szTAdapter)) == 0)
		{
			delete[] pInfoHead;
			return false;
		}
#else
		char *szTAdapter = pAdapterInfo->AdapterName;
#endif
		TCHAR szTAdapterName[256];
		if (!GetAdapterName(szTAdapter, szTAdapterName, LENGTHOF(szTAdapterName)))
		{
			delete[] pInfoHead;
			return false;
		}

		if (pAdapterInfo->Type == MIB_IF_TYPE_ETHERNET &&
			_tcsicmp(szTAdapterName, szAdapterName) == 0)
		{
			if (!szDNS && !IsDHCPEnabled(szTAdapter))
			{
				tcerr << _T("DHCP is not enabled on interface \"") << szAdapterName << _T("\"") << std::endl;
				delete[] pInfoHead;
				return false;
			}

			ULONG ulPerAdapterInfoSize = 0;
			IP_PER_ADAPTER_INFO *pPerAdapterInfoHead = NULL;

			if (GetPerAdapterInfo(pAdapterInfo->Index, NULL, &ulPerAdapterInfoSize) != ERROR_BUFFER_OVERFLOW)
			{
				delete[] pInfoHead;
				return false;
			}
				
			IP_PER_ADAPTER_INFO *pPerAdapterInfo = (IP_PER_ADAPTER_INFO *)new char[ulPerAdapterInfoSize];
			pPerAdapterInfoHead = pPerAdapterInfo;

			if (GetPerAdapterInfo(pAdapterInfo->Index, pPerAdapterInfo, &ulPerAdapterInfoSize) != ERROR_SUCCESS)
			{
				delete[] pInfoHead;
				delete[] pPerAdapterInfoHead;
				return false;
			}

			tcout << _T("previous DNS address(es): ");

			IP_ADDR_STRING *pAddrStr = &pPerAdapterInfo->DnsServerList;
			bool first = true;

			while (pAddrStr)   
			{   
				if (!first)
					tcout << " , ";
				std::cout << pAddrStr->IpAddress.String;
				pAddrStr = pAddrStr->Next;
				first = false;
			}

			std::cout << std::endl;

			tcout << _T("setting DNS to ") << (szDNS ? szDNS : _T("DHCP")) << _T("...");

			bool res = RegSetDNS(szTAdapter, szDNS) && NotifyIPChange(szTAdapter) && FlushDNS();	

			tcout << (res ? _T("success!") : _T("failure!")) << std::endl;

			delete[] pInfoHead;
			delete[] pPerAdapterInfoHead;

			return res;
		}

		pAdapterInfo = pAdapterInfo->Next;
	} while (pAdapterInfo);

	return false;
}

int _tmain(int argc, _TCHAR* argv[])
{
	int dhcp_flag = 0;
	int static_flag = 0;
	int help_flag = 0;
	int c;
	TCHAR *szAddress = NULL;
	TCHAR *szAdapterName = NULL;

	static struct option long_options[] = {
		{ _T("help"),       no_argument,		&help_flag,		1 },
		{ _T("dhcp"),       no_argument,		&dhcp_flag,		1 },
		{ _T("static"),     required_argument,	&static_flag,	1 },
		{ 0, 0, 0, 0 }
    };

	while (1)
	{
		int option_index = 0;
		if ((c = getopt_long(argc, argv, _T("ds:"), long_options, &option_index)) == -1)
			break;

		switch (c)
		{
		case 0:
			if (long_options[option_index].flag == &static_flag)
			{
				szAddress = optarg;
			}
			break;
		case _T('d'):
			dhcp_flag = 1;
			break;
		case _T('s'):
			static_flag = 1;
			szAddress = optarg;
			break;
		case _T('?'):
			usage();
			return 1;
		}
	}

	if (help_flag)
	{
		usage();
		return 0;
	}

	if (optind < argc - 1)
	{
		tcerr << _T("too many arguments.") << std::endl;
		usage();
		return 1;
	}

	if (optind >= argc)
	{
		tcerr << _T("missing interface name.") << std::endl;
		usage();
		return 1;
	}

	szAdapterName = argv[optind];

	if (dhcp_flag == static_flag)
	{
		tcerr << _T("you have to set either --dhcp or --static.") << std::endl;
		return 1;
	}

	if (szAddress && !IsValidIp(szAddress))
	{
		tcerr << _T("invalid IP address: ") << szAddress << std::endl;
		return 1;
	}

	return SetDns(szAdapterName, szAddress)
		? 0
		: 1;
}

